import torch
from gaussian_splatting import GaussianModel, CameraTrainableGaussianModel
from gaussian_splatting.dataset import CameraDataset
from gaussian_splatting.dataset.colmap import colmap_init
from gaussian_splatting.trainer import AbstractTrainer
from gaussian_splatting.trainer.extensions import ScaleRegularizeTrainerWrapper
from lapisgs.dataset import colmap_concat
from lapisgs.trainer.extensions.reduced_3dgs import *


def prepare_gaussians(sh_degree: int, source: str, device: str, trainable_camera: bool = False, load_ply: str = None) -> GaussianModel:
    from reduced_3dgs.prepare import prepare_gaussians as legacy_prepare_gaussians
    gaussians = legacy_prepare_gaussians(sh_degree=sh_degree, source=source, device=device, trainable_camera=trainable_camera, load_ply=load_ply)
    fixed_size = None
    if load_ply:
        fixed_size = gaussians.get_xyz.shape[0]
        colmap_concat(gaussians, source)
        torch.cuda.empty_cache()
    return gaussians, fixed_size


modes = {
    "base": LapisFullTrainer,
    "camera": CameraLapisFullTrainer,
    "shculling": SHCullingLapisFullTrainer,
    "camera-shculling": CameraSHCullingLapisFullTrainer,
}


def prepare_trainer(gaussians: GaussianModel, dataset: CameraDataset, mode: str, load_ply: str = None, with_scale_reg=False, configs={}) -> AbstractTrainer:
    if not load_ply:
        from reduced_3dgs.prepare import prepare_trainer as legacy_prepare_trainer
        modemap = {
            "base": "densify-pruning",
            "camera": "camera-densify-pruning",
            "shculling": "densify-prune-shculling",
            "camera-shculling": "camera-densify-prune-shculling",
        }
        trainer, _ = legacy_prepare_trainer(
            gaussians=gaussians, dataset=dataset, mode=modemap[mode], with_scale_reg=with_scale_reg, quantize=False, configs=configs)
        return trainer
    constructor = modes[mode]
    if with_scale_reg:
        constructor = lambda *args, **kwargs: ScaleRegularizeTrainerWrapper(modes[mode], *args, **kwargs)
    trainer = constructor(
        gaussians,
        scene_extent=dataset.scene_extent(),
        dataset=dataset,
        **configs
    )
    return trainer



from typing import Callable
from gaussian_splatting import GaussianModel
from gaussian_splatting.trainer import DensificationTrainer
from gaussian_splatting.trainer.densifier import AbstractDensifier, SplitCloneDensifier


class PartialDensificationTrainer(DensificationTrainer):

    def __init__(
            self, model: GaussianModel,
            scene_extent: float,
            densifier: AbstractDensifier,
            *args,
            fixed_size=None,
            fix_xyz=True,
            fix_features_dc=True,
            fix_features_rest=True,
            fix_scaling=True,
            fix_rotation=True,
            fix_opacity=False,
            **kwargs
    ):
        super().__init__(model, scene_extent, densifier, *args, **kwargs)
        self.size_fixed_gs = fixed_size if isinstance(fixed_size, int) else model.get_xyz.shape[0]
        self.fix_xyz = fix_xyz
        self.fix_features_dc = fix_features_dc
        self.fix_features_rest = fix_features_rest
        self.fix_scaling = fix_scaling
        self.fix_rotation = fix_rotation
        self.fix_opacity = fix_opacity

    def optim_step(self):
        # freeze all other parameters except opacity in first `size_fixed_gs` points
        # https://github.com/nus-vv-streams/lapis-gs/blob/12dcda37ed43838d7407b28675bc26b7364ae431/scene/gaussian_model.py#L322
        if self.model._xyz.grad is not None and self.fix_xyz:  # grad is None when after pruning or densification
            self.model._xyz.grad[:self.size_fixed_gs] = 0
        if self.model._features_dc.grad is not None and self.fix_features_dc:
            self.model._features_dc.grad[:self.size_fixed_gs] = 0
        if self.model._features_rest.grad is not None and self.fix_features_rest:
            self.model._features_rest.grad[:self.size_fixed_gs] = 0
        if self.model._scaling.grad is not None and self.fix_scaling:
            self.model._scaling.grad[:self.size_fixed_gs] = 0
        if self.model._rotation.grad is not None and self.fix_rotation:
            self.model._rotation.grad[:self.size_fixed_gs] = 0
        if self.model._opacity.grad is not None and self.fix_opacity:
            self.model._opacity.grad[:self.size_fixed_gs] = 0
        return super().optim_step()

    def remove_points(self, rm_mask):
        rm_mask[:self.size_fixed_gs] = False
        return super().remove_points(rm_mask)


def SplitClonePartialDensifierTrainerWrapper(
        noargs_base_densifier_constructor: Callable[[GaussianModel, float], AbstractDensifier],
        model: GaussianModel,
        scene_extent: float,
        *args,
        densify_from_iter=500,
        densify_until_iter=15000,
        densify_interval=100,
        densify_grad_threshold=0.0002,
        densify_percent_dense=0.01,
        densify_percent_too_big=0.8,
        fixed_size=None,
        fix_xyz=True,
        fix_features_dc=True,
        fix_features_rest=True,
        fix_scaling=True,
        fix_rotation=True,
        fix_opacity=False,
        **kwargs):
    # https://github.com/yindaheng98/gaussian-splatting/blob/56576b647d9c5bd05300f5640cd03a8c75a760bc/gaussian_splatting/trainer/densifier/densifier.py#L149
    densifier = noargs_base_densifier_constructor(model, scene_extent)
    densifier = SplitCloneDensifier(
        densifier,
        scene_extent,
        densify_from_iter=densify_from_iter,
        densify_until_iter=densify_until_iter,
        densify_interval=densify_interval,
        densify_grad_threshold=densify_grad_threshold,
        densify_percent_dense=densify_percent_dense,
        densify_percent_too_big=densify_percent_too_big
    )
    return PartialDensificationTrainer(
        model, scene_extent,
        densifier,
        *args,
        fixed_size=fixed_size,
        fix_xyz=fix_xyz,
        fix_features_dc=fix_features_dc,
        fix_features_rest=fix_features_rest,
        fix_scaling=fix_scaling,
        fix_rotation=fix_rotation,
        fix_opacity=fix_opacity,
        **kwargs
    )



import torch
import torch.nn as nn
from gaussian_splatting import GaussianModel
from gaussian_splatting.camera import build_camera
from gaussian_splatting.dataset.colmap import read_colmap_cameras, ColmapCameraDataset, colmap_init
from .dataset import RescaleCameraDatasetIface, RescaleTrainableCameraDataset


class RescaleColmapCameraDataset(RescaleCameraDatasetIface, ColmapCameraDataset):
    def __init__(self, colmap_folder, load_depth=False, rescale_factor=1.0):
        # https://github.com/yindaheng98/gaussian-splatting/blob/56576b647d9c5bd05300f5640cd03a8c75a760bc/gaussian_splatting/dataset/colmap/dataset.py#L89
        self.raw_cameras = read_colmap_cameras(colmap_folder, load_depth=load_depth)
        self.raw_image_hw = [(raw_camera.image_height, raw_camera.image_width) for raw_camera in self.raw_cameras]
        self.raw_cameras = [
            raw_camera._replace(
                image_height=round(h*rescale_factor), image_width=round(w*rescale_factor),
            ) for (h, w), raw_camera in zip(self.raw_image_hw, self.raw_cameras)]
        self.cameras = [build_camera(**cam._asdict(), custom_data=dict(fullimage_width=w, fullimage_height=h)) for (h, w), cam in zip(self.raw_image_hw, self.raw_cameras)]

    def to(self, device):
        # https://github.com/yindaheng98/gaussian-splatting/blob/56576b647d9c5bd05300f5640cd03a8c75a760bc/gaussian_splatting/dataset/colmap/dataset.py#L94
        self.cameras = [build_camera(**cam._asdict(), custom_data=dict(fullimage_width=w, fullimage_height=h), device=device) for (h, w), cam in zip(self.raw_image_hw, self.raw_cameras)]
        return self


def RescaleColmapTrainableCameraDataset(colmap_folder, load_depth=False, rescale_factor=1.0):
    # https://github.com/yindaheng98/gaussian-splatting/blob/56576b647d9c5bd05300f5640cd03a8c75a760bc/gaussian_splatting/dataset/colmap/dataset.py#L105
    return RescaleTrainableCameraDataset(RescaleColmapCameraDataset(colmap_folder, load_depth=load_depth, rescale_factor=rescale_factor))


def colmap_concat(model: GaussianModel, colmap_folder: str):
    temp = GaussianModel(sh_degree=model.max_sh_degree).to(model._xyz.device)
    colmap_init(model=temp, colmap_folder=colmap_folder)
    model.update_points_add(
        xyz=nn.Parameter(torch.cat([model._xyz.detach(), temp._xyz.detach()], dim=0)),
        features_dc=nn.Parameter(torch.cat([model._features_dc.detach(), temp._features_dc.detach()], dim=0)),
        features_rest=nn.Parameter(torch.cat([model._features_rest.detach(), temp._features_rest.detach()], dim=0)),
        scaling=nn.Parameter(torch.cat([model._scaling.detach(), temp._scaling.detach()], dim=0)),
        rotation=nn.Parameter(torch.cat([model._rotation.detach(), temp._rotation.detach()], dim=0)),
        opacity=nn.Parameter(torch.cat([model._opacity.detach(), temp._opacity.detach()], dim=0)),
    )
    return model



import os
from typing import Tuple
import torch
from gaussian_splatting import GaussianModel
from gaussian_splatting.dataset import CameraDataset
from gaussian_splatting.trainer import AbstractTrainer
from gaussian_splatting.train import save_cfg_args, training
from lapisgs.prepare import prepare_dataset
from lapisgs.prepare_reduced import modes, prepare_gaussians, prepare_trainer


def prepare_training(sh_degree: int, source: str, device: str, mode: str, trainable_camera: bool = False, load_ply: str = None, load_camera: str = None, load_depth=False, rescale_factor=1.0, with_scale_reg=False, configs={}) -> Tuple[CameraDataset, GaussianModel, AbstractTrainer]:
    dataset = prepare_dataset(source=source, device=device, trainable_camera=trainable_camera, load_camera=load_camera, load_depth=load_depth, rescale_factor=rescale_factor)
    gaussians, fixed_size = prepare_gaussians(sh_degree=sh_degree, source=source, device=device, trainable_camera=trainable_camera, load_ply=load_ply)
    if isinstance(fixed_size, int):
        configs["fixed_opacity_size"] = configs["fixed_size"] = fixed_size
    trainer = prepare_trainer(gaussians=gaussians, dataset=dataset, mode=mode, load_ply=load_ply, with_scale_reg=with_scale_reg, configs=configs)
    return dataset, gaussians, trainer


if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("--sh_degree", default=3, type=int)
    parser.add_argument("-s", "--source", required=True, type=str)
    parser.add_argument("-d", "--destination", required=True, type=str)
    parser.add_argument("-i", "--iteration", default=30000, type=int)
    parser.add_argument("-l", "--foundation_gs_path", default=None, type=str)
    parser.add_argument("--load_camera", default=None, type=str)
    parser.add_argument("--no_depth_data", action="store_true")
    parser.add_argument("--rescale_factor", default=1.0, type=float)
    parser.add_argument("--with_scale_reg", action="store_true")
    parser.add_argument("--mode", choices=sorted(list(modes.keys())), default="base")
    parser.add_argument("--save_iterations", nargs="+", type=int, default=[7000, 30000])
    parser.add_argument("--device", default="cuda", type=str)
    parser.add_argument("-o", "--option", default=[], action='append', type=str)
    args = parser.parse_args()
    save_cfg_args(args.destination, args.sh_degree, args.source)
    torch.autograd.set_detect_anomaly(False)

    configs = {o.split("=", 1)[0]: eval(o.split("=", 1)[1]) for o in args.option}
    dataset, gaussians, trainer = prepare_training(
        sh_degree=args.sh_degree, source=args.source, device=args.device, mode=args.mode, trainable_camera="camera" in args.mode,
        load_ply=args.foundation_gs_path, load_camera=args.load_camera, load_depth=not args.no_depth_data, rescale_factor=args.rescale_factor, with_scale_reg=args.with_scale_reg, configs=configs)
    dataset.save_cameras(os.path.join(args.destination, "cameras.json"))
    torch.cuda.empty_cache()
    training(
        dataset=dataset, gaussians=gaussians, trainer=trainer,
        destination=args.destination, iteration=args.iteration, save_iterations=args.save_iterations,
        device=args.device)



import os
import torch
from gaussian_splatting.train import save_cfg_args, training
from lapisgs.prepare_reduced import modes
from lapisgs.train_reduced import prepare_training


if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("--sh_degree", default=3, type=int)
    parser.add_argument("-s", "--source", required=True, type=str)
    parser.add_argument("-d", "--destination", required=True, type=str)
    parser.add_argument("-i", "--iteration", default=30000, type=int)
    parser.add_argument("--load_camera", default=None, type=str)
    parser.add_argument("--no_depth_data", action="store_true")
    parser.add_argument("--resolution_scales", action="append", type=int, default=[8, 4, 2, 1])
    parser.add_argument("--with_scale_reg", action="store_true")
    parser.add_argument("--mode", choices=sorted(list(modes.keys())), default="base")
    parser.add_argument("--save_iterations", nargs="+", type=int, default=[7000, 30000])
    parser.add_argument("--device", default="cuda", type=str)
    parser.add_argument("-o", "--option", default=[], action='append', type=str)
    args = parser.parse_args()
    torch.autograd.set_detect_anomaly(False)

    load_camera = None
    foundation_gs_path = None
    for resolution in args.resolution_scales:
        destination = os.path.join(args.destination, f"{resolution}x")
        save_cfg_args(destination, args.sh_degree, args.source)

        configs = {o.split("=", 1)[0]: eval(o.split("=", 1)[1]) for o in args.option}
        dataset, gaussians, trainer = prepare_training(
            sh_degree=args.sh_degree, source=args.source, device=args.device, mode=args.mode, trainable_camera="camera" in args.mode,
            load_ply=foundation_gs_path, load_camera=load_camera, load_depth=not args.no_depth_data, rescale_factor=1/resolution, with_scale_reg=args.with_scale_reg, configs=configs)
        dataset.save_cameras(os.path.join(destination, "cameras.json"))
        torch.cuda.empty_cache()
        training(
            dataset=dataset, gaussians=gaussians, trainer=trainer,
            destination=destination, iteration=args.iteration, save_iterations=args.save_iterations,
            device=args.device)
        load_camera = os.path.join(destination, "cameras.json")
        foundation_gs_path = os.path.join(destination, "point_cloud", f"iteration_{args.iteration}", "point_cloud.ply")



from .dataset import RescaleJSONCameraDataset, RescaleTrainableCameraDataset, RescaleFixedTrainableCameraDataset
from .colmap import RescaleColmapCameraDataset, RescaleColmapTrainableCameraDataset, colmap_concat



from .trainer import LapisReducedTrainer, BaseLapisReducedTrainer
from .trainer import PartialPrunerInDensifyTrainerWrapper, PartialPrunerInDensifyTrainer, DepthPrunerInDensifyTrainer
from .importance import LapisImportanceTrainer, BaseLapisImportanceTrainer
from .importance import PartialImportancePrunerInDensifyTrainer, DepthPartialImportancePrunerInDensifyTrainer
from .combinations import LapisFullTrainer, CameraLapisFullTrainer
from .combinations import SHCullingLapisTrainer, SHCullingLapisFullTrainer
from .combinations import CameraSHCullingLapisTrainer, CameraSHCullingLapisFullTrainer
